/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/** STM32F1 example
 *  @file main.c
 *  @author King Kévin <kingkevin@cuvoodoo.info>
 *  @date 2016
 */

/* standard libraries */
#include <stdint.h> // standard integer types
#include <stdbool.h> // boolean type
#include <string.h> // string utilities

/* STM32 (including CM3) libraries */
#include <libopencmsis/core_cm3.h> // Cortex M3 utilities
#include <libopencm3/cm3/scb.h> // vector table definition
#include <libopencm3/cm3/nvic.h> // interrupt utilities
#include <libopencm3/stm32/gpio.h> // general purpose input output library
#include <libopencm3/stm32/rcc.h> // real-time control clock library
#include <libopencm3/stm32/exti.h> // external interrupt utilities
#include <libopencm3/stm32/rtc.h> // real time clock utilities
#include <libopencm3/stm32/iwdg.h> // independent watchdog utilities
#include <libopencm3/stm32/dbgmcu.h> // debug utilities
#include <libopencm3/stm32/flash.h> // flash utilities
#include <libopencm3/stm32/timer.h> // timer utilities

/* own libraries */
#include "global.h" // board definitions
#include "print.h" // printing utilities
#include "usart.h" // USART utilities
#include "usb_cdcacm.h" // USB CDC ACM utilities
#include "rtc_ds1307.h" // DS1307 RTC utilities
#include "led_tm1637.h" // TM1637 7-segment controller utilities
#include "led_max7219.h" // MAX7219 7-segment controller utilities

#define WATCHDOG_PERIOD 10000 /**< watchdog period in ms */

/** @defgroup main_flags flag set in interrupts to be processed in main task
 *  @{
 */
volatile bool rtc_tick_flag = false; /**< flag set when RTC ticked */
volatile bool frame_flag = false; /**< flag set when a frame has passed */
/** @} */

#define SQUARE_WAVE_PORT B /**< port connected to RTC DS1307 square wave output */
#define SQUARE_WAVE_PIN 0 /**< pin connected to RTC DS1307 square wave output */
volatile uint8_t rtc_seconds = 0; /**< number of seconds passed incremented by the square wave */

#define STANDBY_TIMEOUT 30  /**< number of seconds after last shake before going down */
volatile uint16_t standby_timer = 0; /**< number of seconds since last wake-up/activity */

#define FRAME_TIMER 1 /**< timer to count frame time */
#define FRAME_RATE 25 /**< frame rate */
volatile uint8_t frame_count = 0; /**< number of frames passed */

/** user input command */
static char command[32] = {0};
/** user input command index */
uint8_t command_i = 0;

size_t putc(char c)
{
	size_t length = 0; // number of characters printed
	static char newline = 0; // to remember on which character we sent the newline
	if (0==c) {
		length = 0; // don't print string termination character
	} else if ('\r' == c || '\n' == c) { // send CR+LF newline for most carriage return and line feed combination
		if (0==newline || c==newline) { // send newline only if not already send (and only once on \r\n or \n\r)
				usart_putchar_nonblocking('\r'); // send CR over USART
				cdcacm_putchar('\r'); // send CR over USB
				usart_putchar_nonblocking('\n'); // send LF over USART
				cdcacm_putchar('\n'); // send LF over USB
				length += 2; // remember we printed 2 characters
				newline = c; // remember on which character we sent the newline
		} else {
			length = 0; // the \r or \n of \n\r or \r\n has already been printed
		}
	} else {
		usart_putchar_nonblocking(c); // send byte over USART
		cdcacm_putchar(c); // send byte over USB
		newline = 0; // clear new line
		length++; // remember we printed 1 character
	}
	return length; // return number of characters printed	
}

/** process user command
 *  @param[in] str user command string (\0 ended)
 */
static void process_command(char* str)
{
	// split command
	const char* delimiter = " ";
	char* word = strtok(str,delimiter);
	if (!word) {
		goto error;
	}
	// parse command
	if (0==strcmp(word,"help")) {
		printf("available commands:\n");
		printf("led [on|off|toggle]\n");
		printf("time [HH:MM:SS]\n");
		printf("date [YYYY-MM-DD]\n");
	} else if (0==strcmp(word,"led")) {
		word = strtok(NULL,delimiter);
		if (!word) {	
			goto error;
		} else if (0==strcmp(word,"on")) {
			led_on(); // switch LED on
			printf("LED switched on\n"); // notify user		
		} else if (0==strcmp(word,"off")) {
			led_off(); // switch LED off
			printf("LED switched off\n"); // notify user
		} else if (0==strcmp(word,"toggle")) {
			led_toggle(); // toggle LED
			printf("LED toggled\n"); // notify user
		} else {
			goto error;
		}
	} else if (0==strcmp(word,"time")) {
		word = strtok(NULL,delimiter);
		if (!word) {
			printf("current time: %02u:%02u:%02u\n", rtc_ds1307_read_hours(), rtc_ds1307_read_minutes(), rtc_ds1307_read_seconds()); // get and print time from external RTC
		} else if (strlen(word)!=8 || word[0]<'0' || word[0]>'2' || word[1]<'0' || word[1]>'9' || word[3]<'0' || word[3]>'5' || word[4]<'0' || word[4]>'9' || word[6]<'0' || word[6]>'5' || word[7]<'0' || word[7]>'9') { // time format is incorrect
				goto error;
		} else {
			if (!rtc_ds1307_write_hours((word[0]-'0')*10+(word[1]-'0')*1)) {
				printf("setting hours failed\n");
			} else if (!rtc_ds1307_write_minutes((word[3]-'0')*10+(word[4]-'0')*1)) {
				printf("setting minutes failed\n");
			} else if (!rtc_ds1307_write_seconds((word[6]-'0')*10+(word[7]-'0')*1)) {
				printf("setting seconds failed\n");
			} else {
				rtc_ds1307_oscillator_enable(); // be sure the oscillation is enabled
				printf("time set\n");
			}
		}
	} else if (0==strcmp(word,"date")) {
		word = strtok(NULL,delimiter);
		if (!word) {
			printf("current date: 20%02u-%02u-%02u\n", rtc_ds1307_read_year(), rtc_ds1307_read_month(), rtc_ds1307_read_date());
		} else if (strlen(word)!=10 || word[0]!='2' || word[1]!='0' || word[2]<'0' || word[2]>'9' || word[3]<'0' || word[3]>'9' || word[5]<'0' || word[5]>'1' || word[6]<'0' || word[6]>'9' || word[8]<'0' || word[8]>'3' || word[9]<'0' || word[9]>'9') {
				goto error;
		} else {
			if (!rtc_ds1307_write_year((word[2]-'0')*10+(word[3]-'0')*1)) {
				printf("setting year failed\n");
			} else if (!rtc_ds1307_write_month((word[5]-'0')*10+(word[6]-'0')*1)) {
				printf("setting month failed\n");
			} else if (!rtc_ds1307_write_date((word[8]-'0')*10+(word[9]-'0')*1)) {
				printf("setting day failed\n");
			} else {
				printf("date set\n");
			}
		}
	} else {
		goto error;
	}

	return; // command successfully processed
error:
	printf("command not recognized. enter help to list commands\n");
	return;
}

/** program entry point
 *  this is the firmware function started by the micro-controller
 */
void main(void);
void main(void)
{	
	rcc_clock_setup_in_hse_8mhz_out_72mhz(); // use 8 MHz high speed external clock to generate 72 MHz internal clock

#if DEBUG
	// enable functionalities for easier debug
	DBGMCU_CR |= DBGMCU_CR_IWDG_STOP; // stop independent watchdog counter when code is halted
	DBGMCU_CR |= DBGMCU_CR_WWDG_STOP; // stop window watchdog counter when code is halted
	DBGMCU_CR |= DBGMCU_CR_STANDBY; // allow debug also in standby mode (keep digital part and clock powered)
	DBGMCU_CR |= DBGMCU_CR_STOP; // allow debug also in stop mode (keep clock powered)
	DBGMCU_CR |= DBGMCU_CR_SLEEP; // allow debug also in sleep mode (keep clock powered)
#else
	// setup watchdog to reset in case we get stuck (i.e. when an error occurred)
	iwdg_set_period_ms(WATCHDOG_PERIOD); // set independent watchdog period
	iwdg_start(); // start independent watchdog
#endif

	board_setup(); // setup board
    usart_setup(); // setup USART for user communication
	cdcacm_setup(); // setup USB ACM (serial) for user communication
	printf("welcome to the STM32F1 CuVoodoo clapperboard\n"); // print welcome message

#if !(DEBUG)
	// show watchdog information
	printf("watchdog set to (%2u.%2us)\n",WATCHDOG_PERIOD/1000, (WATCHDOG_PERIOD/10)%100);
	if (FLASH_OBR&FLASH_OBR_OPTERR) {
		printf("option bytes not set in flash: software wachtdog used (not started at reset)\n");
	} else if (FLASH_OBR&FLASH_OBR_WDG_SW) {
		printf("software wachtdog used (not started at reset)\n");
	} else {
		printf("hardware wachtdog used (started at reset)\n");
	}
#endif

	// enable wake up
	rcc_periph_clock_enable(RCC_PWR); // enable power domain clock
	if (pwr_get_wakeup_flag()) {
		printf("woke up from standby\n");
	}	
	pwr_enable_wakeup_pin();

	// disable internal RTC
	printf("disable internal RTC: ");
	rcc_periph_clock_enable(RCC_PWR); // enable power domain clock
	rcc_periph_clock_enable(RCC_BKP); // enable backup domain clock
	pwr_disable_backup_domain_write_protect(); // enable write on backup domain (including RTC)
	RCC_BDCR &= ~RCC_BDCR_RTCEN; // disable RTC
	pwr_enable_backup_domain_write_protect(); // re-enable write protect
	printf("OK\n");

	// setup external RTC
	printf("setup external RTC: ");
	rtc_ds1307_setup(); // setup external RTC module
	// enable square wave output and configure input interrupt
	rtc_ds1307_write_square_wave(1); // enable 1 Hz square wave output to sync on seconds
	rcc_periph_clock_enable(RCC_GPIO(SQUARE_WAVE_PORT)); // enable clock for GPIO
	gpio_set_mode(GPIO(SQUARE_WAVE_PORT), GPIO_MODE_INPUT, GPIO_CNF_INPUT_FLOAT, GPIO(SQUARE_WAVE_PIN)); // set button pin to input
	rcc_periph_clock_enable(RCC_AFIO); // enable alternate function clock for external interrupt
	exti_select_source(EXTI(SQUARE_WAVE_PIN), GPIO(SQUARE_WAVE_PORT)); // mask external interrupt of this pin only for this port
	exti_set_trigger(EXTI(SQUARE_WAVE_PIN), EXTI_TRIGGER_FALLING); // trigger on falling edge of square wave (this is also when the RTC register are updated)
	exti_enable_request(EXTI(SQUARE_WAVE_PIN)); // enable external interrupt
	nvic_enable_irq(NVIC_EXTI_IRQ(SQUARE_WAVE_PIN)); // enable interrupt
	printf("OK\n");

	// display date
	uint8_t* rtc_ds1307_time = rtc_ds1307_read_time(); // get time/date from external RTC
	if (rtc_ds1307_time==NULL) {
		printf("could not get time from DS1307\n");
	} else {
		rtc_seconds = rtc_ds1307_time[0]; // remember seconds of minute
		printf("current date: 20%02u-%02u-%02u %02u:%02u:%02u\n", rtc_ds1307_time[6], rtc_ds1307_time[5], rtc_ds1307_time[4], rtc_ds1307_time[2], rtc_ds1307_time[1], rtc_ds1307_time[0]);
	}
	// verify is external RTC is running
	if (rtc_ds1307_oscillator_disabled()) {
		printf("/!\\ RTC oscillator is disabled: the battery may be empty\n");
		rtc_ds1307_oscillator_enable(); // enable oscillator again
	}

	// setup TM1637 and MAX7219 7-segments displays
	printf("setup 7-segment displays: ");
	led_tm1637_setup(); // setup TM1637
	led_max7219_setup(2); // setup MAX7219
	if (!led_tm1637_time(88,88)) { // test TM1637 display
		printf("could not send time to TM1637\n");
	}
	if (!led_tm1637_on()) { // switch on TM1637 display
		printf("could not switch on TM1637\n");
	}
	led_max7219_intensity(15,8,0xff); // set brightness max and enable all digits
	led_max7219_test(true,0xff); // test all MAX7219 displays
	for (uint32_t i=0; i<5000000; i++) { // wait a bit to have the user check the display
		__asm__("nop");
	}
	if (!led_tm1637_text("    ")) { // clear display
		printf("could not clear\n");
	}
	if (!led_tm1637_off()) { // switch off display
		printf("could not switch off TM1637\n");
	}
	led_max7219_test(false,0xff); // go back in normal operation
	led_max7219_off(0xff); // switch displays off
	printf("OK\n");

	// display date and time on 7-segments	
	led_max7219_number(20000000+rtc_ds1307_time[6]*10000+rtc_ds1307_time[5]*100+rtc_ds1307_time[4], 0x14, 1); // display date on 2nd display
	led_max7219_number(rtc_ds1307_time[2]*1000000+rtc_ds1307_time[1]*10000+rtc_ds1307_time[0]*100, 0x54, 0); // display time on 1nd display
	led_max7219_on(0xff); // switch displays on

	// setup frame timer
	printf("setup frame timer: ");
	rcc_periph_clock_enable(RCC_TIM(FRAME_TIMER)); // enable clock for timer block
	timer_reset(TIM(FRAME_TIMER)); // reset timer state
	timer_set_mode(TIM(FRAME_TIMER), TIM_CR1_CKD_CK_INT, TIM_CR1_CMS_EDGE, TIM_CR1_DIR_UP); // set timer mode, use undivided timer clock, edge alignment (simple count), and count up
	timer_set_prescaler(TIM(FRAME_TIMER), (rcc_ahb_frequency/0xffff+1)-1); // set the prescaler to so count up to one second
	timer_set_period(TIM(FRAME_TIMER), 0xffff/FRAME_RATE); // overflow at the end of every rate
	timer_update_on_overflow(TIM(FRAME_TIMER)); // only use counter overflow as UEV source (use overflow as timeout)
	timer_clear_flag(TIM(FRAME_TIMER), TIM_SR_UIF); // clear flag
	timer_enable_irq(TIM(FRAME_TIMER), TIM_DIER_UIE); // enable update interrupt for timer
#if FRAME_TIMER==1
	nvic_enable_irq(NVIC_TIM1_UP_IRQ); // catch interrupt in service routine
#else
	nvic_enable_irq(NVIC_TIM_IRQ(FRAME_TIMER)); // catch interrupt in service routine
#endif
	timer_set_counter(TIM(FRAME_TIMER),0); // restart timer
	timer_enable_counter(TIM(FRAME_TIMER)); // enable timer to start counting frames
	printf("OK\n");

	// main loop
	printf("command input: ready\n");
	bool action = false; // if an action has been performed don't go to sleep
	button_flag = false; // reset button flag
	char c = '\0'; // to store received character
	bool char_flag = false; // a new character has been received
	while (true) { // infinite loop
		iwdg_reset(); // kick the dog
		while (usart_received) { // data received over UART
			action = true; // action has been performed
			led_toggle(); // toggle LED
			c = usart_getchar(); // store receive character
			char_flag = true; // notify character has been received
		}
		while (cdcacm_received) { // data received over USB
			action = true; // action has been performed
			led_toggle(); // toggle LED
			c = cdcacm_getchar(); // store receive character
			char_flag = true; // notify character has been received
		}
		while (char_flag) { // user data received
			char_flag = false; // reset flag
			action = true; // action has been performed
			putc(c); // echo receive character
			if (c=='\r' || c=='\n') { // end of command received
				if (command_i>0) { // there is a command to process
					command[command_i] = 0; // end string
					command_i = 0; // prepare for next command
					process_command(command); // process user command
				}
			} else { // user command input
				command[command_i] = c; // save command input
				if (command_i<LENGTH(command)-2) { // verify if there is place to save next character
					command_i++; // save next character
				}
			}
		}
		while (button_flag) { // user pressed button
			action = true; // action has been performed
			printf("button pressed\n");
			led_toggle(); // toggle LED
			for (uint32_t i=0; i<1000000; i++) { // wait a bit to remove noise and double trigger
				__asm__("nop");
			}
			button_flag = false; // reset flag
		}
		while (frame_flag) { // a frame has passed
			frame_flag = false; // reset flag
			action = true; // action has been performed
			uint8_t dots = 0x40; // dot for minutes on display
			if (0==(rtc_seconds%2)) { // even seconds
				dots |= 0x10; // add dot for seconds
			}
			if (0==(frame_count%2)) { // even frames
				dots |= 0x04; // add dot for frame
			}
			led_max7219_number(rtc_ds1307_time[2]*1000000+rtc_ds1307_time[1]*10000+rtc_seconds*100+(frame_count%100), dots, 0); // display frame time on 1nd display
		}
		while (rtc_tick_flag) { // the external RTC ticked
			rtc_tick_flag = false; // reset flag
			action = true; // action has been performed
			led_toggle(); // toggle LED (good to indicate if main function is stuck)
			frame_count = 0; // resync frame counter to second
			if (standby_timer>=STANDBY_TIMEOUT) { // standby timeout complete
				// go into standby mode
				printf("standing by\n");
				led_max7219_off(0xff); // switch displays off
				pwr_clear_wakeup_flag(); // clear flag
				SCB_SCR |= SCB_SCR_SLEEPDEEP; // enable deep sleep
				pwr_set_standby_mode(); // go to deep sleep
			}
			if (rtc_seconds>=60) { // one minute passed
				rtc_ds1307_time = rtc_ds1307_read_time(); // get time/date from external RTC
				if (rtc_ds1307_time==NULL) {
					printf("could not get time from DS1307: resetting\n");
					rtc_ds1307_setup(); // resetting periph
				} else {
					rtc_seconds = rtc_ds1307_time[0]; // get actual number of seconds
					if (0==rtc_ds1307_time[1] && 0==rtc_ds1307_time[2]) { // new day arrived
						led_max7219_number(20000000+rtc_ds1307_time[6]*10000+rtc_ds1307_time[5]*100+rtc_ds1307_time[4], 0x14, 1); // display date on 2nd display
					}
				}
			}
		}
		while (pwr_get_wakeup_flag()) { // someone is moving the clapperboard
			pwr_clear_wakeup_flag(); // clear flag
			standby_timer = 0; // restart standby timer
		}
		if (action) { // go to sleep if nothing had to be done, else recheck for activity
			action = false;
		} else {
			__WFI(); // go to sleep and wait for interrupt
		}
	} // main loop
}

/** RTC square wave input ISR to synchronize to seconds and count them */
void EXTI_ISR(SQUARE_WAVE_PIN)(void)
{
    exti_reset_request(EXTI(SQUARE_WAVE_PIN)); // reset interrupt
	rtc_seconds++; // increment number of seconds passed
	if (standby_timer<STANDBY_TIMEOUT) { // timeout countdown not complete
		standby_timer++; // continue counting down
	}
    rtc_tick_flag = true; // let main know a second passed
}

/** frame timer ISR */
#if FRAME_TIMER==1
void tim1_up_isr(void)
#else
void TIM_ISR(FRAME_TIMER)(void)
#endif
{
	if (timer_get_flag(TIM(FRAME_TIMER), TIM_SR_UIF)) { // overflow update event happened
		timer_clear_flag(TIM(FRAME_TIMER), TIM_SR_UIF); // clear flag
		frame_count++; // increment frame counter since frame finished
		frame_flag = true; // let main know the counter changed
	} else { // no other interrupt should occur
		while (true); // unhandled exception: wait for the watchdog to bite
	}
}

